from enum import Enum

class COMMAND_CLASS(Enum):
    NO_OPERATION                             = 0x00
    BASIC                                    = 0x20
    CONTROLLER_REPLICATION                   = 0x21
    APPLICATION_STATUS                       = 0x22
    ZIP_SERVICES                             = 0x23
    ZIP_SERVER                               = 0x24
    SWITCH_BINARY                            = 0x25
    SWITCH_MULTILEVEL                        = 0x26
    SWITCH_MULTILEVEL_V2                     = 0x26
    SWITCH_ALL                               = 0x27
    SWITCH_TOGGLE_BINARY                     = 0x28
    SWITCH_TOGGLE_MULTILEVEL                 = 0x29
    CHIMNEY_FAN                              = 0x2A
    SCENE_ACTIVATION                         = 0x2B
    SCENE_ACTUATOR_CONF                      = 0x2C
    SCENE_CONTROLLER_CONF                    = 0x2D
    ZIP_CLIENT                               = 0x2E
    ZIP_ADV_SERVICES                         = 0x2F
    SENSOR_BINARY                            = 0x30
    SENSOR_MULTILEVEL                        = 0x31
    SENSOR_MULTILEVEL_V2                     = 0x31
    METER                                    = 0x32
    ZIP_ADV_SERVER                           = 0x33
    ZIP_ADV_CLIENT                           = 0x34
    METER_PULSE                              = 0x35
    METER_TBL_CONFIG                         = 0x3C
    METER_TBL_MONITOR                        = 0x3D
    METER_TBL_PUSH                           = 0x3E
    THERMOSTAT_HEATING                       = 0x38
    THERMOSTAT_MODE                          = 0x40
    THERMOSTAT_OPERATING_STATE               = 0x42
    THERMOSTAT_SETPOINT                      = 0x43
    THERMOSTAT_FAN_MODE                      = 0x44
    THERMOSTAT_FAN_STATE                     = 0x45
    CLIMATE_CONTROL_SCHEDULE                 = 0x46
    THERMOSTAT_SETBACK                       = 0x47
    DOOR_LOCK_LOGGING                        = 0x4C
    SCHEDULE_ENTRY_LOCK                      = 0x4E
    BASIC_WINDOW_COVERING                    = 0x50
    MTP_WINDOW_COVERING                      = 0x51
    ASSOCIATION_GRP_INFO                     = 0x59
    DEVICE_RESET_LOCALLY                     = 0x5A
    CENTRAL_SCENE                            = 0x5B
    IP_ASSOCIATION                           = 0x5C
    ANTITHEFT                                = 0x5D
    ZWAVEPLUS_INFO                           = 0x5E
    MULTI_CHANNEL_V2                         = 0x60
    MULTI_INSTANCE                           = 0x60
    DOOR_LOCK                                = 0x62
    USER_CODE                                = 0x63
    BARRIER_OPERATOR                         = 0x66
    CONFIGURATION                            = 0x70
    CONFIGURATION_V2                         = 0x70
    ALARM                                    = 0x71
    MANUFACTURER_SPECIFIC                    = 0x72
    POWERLEVEL                               = 0x73
    PROTECTION                               = 0x75
    PROTECTION_V2                            = 0x75
    LOCK                                     = 0x76
    NODE_NAMING                              = 0x77
    FIRMWARE_UPDATE_MD                       = 0x7A
    GROUPING_NAME                            = 0x7B
    REMOTE_ASSOCIATION_ACTIVATE              = 0x7C
    REMOTE_ASSOCIATION                       = 0x7D
    BATTERY                                  = 0x80
    CLOCK                                    = 0x81
    HAIL                                     = 0x82
    WAKE_UP                                  = 0x84
    WAKE_UP_V2                               = 0x84
    ASSOCIATION                              = 0x85
    ASSOCIATION_V2                           = 0x85
    VERSION                                  = 0x86
    INDICATOR                                = 0x87
    PROPRIETARY                              = 0x88
    LANGUAGE                                 = 0x89
    TIME                                     = 0x8A
    TIME_PARAMETERS                          = 0x8B
    GEOGRAPHIC_LOCATION                      = 0x8C
    COMPOSITE                                = 0x8D
    MULTI_CHANNEL_ASSOCIATION_V2             = 0x8E
    MULTI_INSTANCE_ASSOCIATION               = 0x8E
    MULTI_CMD                                = 0x8F
    ENERGY_PRODUCTION                        = 0x90
    MANUFACTURER_PROPRIETARY                 = 0x91
    SCREEN_MD                                = 0x92
    SCREEN_MD_V2                             = 0x92
    SCREEN_ATTRIBUTES                        = 0x93
    SCREEN_ATTRIBUTES_V2                     = 0x93
    SIMPLE_AV_CONTROL                        = 0x94
    AV_CONTENT_DIRECTORY_MD                  = 0x95
    AV_RENDERER_STATUS                       = 0x96
    AV_CONTENT_SEARCH_MD                     = 0x97
    SECURITY                                 = 0x98
    AV_TAGGING_MD                            = 0x99
    IP_CONFIGURATION                         = 0x9A
    ASSOCIATION_COMMAND_CONFIGURATION        = 0x9B
    SENSOR_ALARM                             = 0x9C
    SILENCE_ALARM                            = 0x9D
    SENSOR_CONFIGURATION                     = 0x9E
    MARK                                     = 0xEF
    NON_INTEROPERABLE                        = 0xF0



import platform
import sys,time
import logging

# Zwave imports
from openzwave.option import ZWaveOption
from openzwave.network import ZWaveNetwork
from openzwave.object import ZWaveException
from pydispatch import dispatcher

from xaal.lib import tools,Device
from . import cmdclass
from . import products

from prettytable import PrettyTable
import atexit

PACKAGE_NAME = "xaal.zwave"
logger = logging.getLogger(PACKAGE_NAME)

logging.getLogger("libopenzwave").setLevel(logging.INFO)
logging.getLogger("openzwave").setLevel(logging.INFO)

class GW(object):
    
    def __init__(self,engine):
        self.products = []
        self.engine = engine
        self.config()
        atexit.register(self._exit)
        self.setup_gw()
        self.setup_network()
        self.setup_products()

    def config(self):
        self.save_config = False
        cfg = tools.load_cfg(PACKAGE_NAME)
        if not cfg:
            logger.info('Missing config file, building a new one')
            cfg = tools.new_cfg(PACKAGE_NAME)
            cfg['config']['port'] = '/dev/zwave'
            cfg['products'] = {}
            cfg.write()
        self.cfg = cfg
        
    def update_value(self,network,node,value):
        dev = self.get_product(node.node_id)
        if dev:
            dev.handle_value_changed(value=value)
        else:
            #print('update_value: {}.'.format(value))
            pass
            
    def update_node(self,network,node):
        if self.get_product(node.node_id) != None: return
        if node.is_ready:
            logger.info("new zwave device [%d]: %s" % (node.node_id,node.product_name))
            dev = self.add_product(node.node_id)
        
    def setup_network(self):
        cfg = self.cfg['config']
        options = ZWaveOption(cfg['port'])
        options.set_console_output(False)
        # comments the following lines if you want a log file.
        options.set_save_log_level('None')
        options.set_log_file("/dev/null")
        options.lock()
        self.network = ZWaveNetwork(options, autostart=False)
        dispatcher.connect(self.update_node,ZWaveNetwork.SIGNAL_NODE)    
        dispatcher.connect(self.update_value,ZWaveNetwork.SIGNAL_VALUE)
        # searching for device w/ wait loop is mandatory. If you don't do that
        # the openzwave lib will segfault.
        self.network.start()
        logger.info('searching for Zwave devices, this may take a while')
        for i in range(0,60):
            if self.network.is_ready:
                logger.info('Zwave network ready')
                break
            else:
                time.sleep(1.0)
                print("\rLoading [%d] " % i,end='')
                
    def setup_products(self):
        for k in self.network.nodes:
            node = self.network.nodes[k]
            if self.get_product(node.node_id):
                continue
            if (node and node.is_ready):
                self.add_product(k)

    def get_product(self,node_id):
        for dev in self.products:
            if dev.node.node_id == node_id:
                return dev

    def add_product(self,node_id):
        node = self.network.nodes[node_id]
        sig = "%s:%s:%s" % (node.manufacturer_id,node.product_id,node.product_type)
        klass=products.search(sig)
        if klass:
            dev = klass(node,self)
            self.products.append(dev)
            self.engine.add_devices(dev.devices)
            self.update_embedded()
            dev.start()
        else:
            logger.info('Unknow device %s %s' % (node.product_name,sig))
        
    def setup_gw(self):
        # last step build the GW device
        gw            = Device("gateway.basic")
        gw.address    = tools.get_uuid(self.cfg['config']['addr'])
        gw.vendor_id  = "IHSEV"
        gw.product_id = "OpenZwave gateway"
        gw.version    = 0.1
        gw.url        = "http://www.openzwave.com/"
        gw.info       = "%s@%s" % (PACKAGE_NAME,platform.node())
        emb = gw.new_attribute('embedded',[])
        self.engine.add_device(gw)
        self.gw = gw
        
    def update_embedded(self):
        result = [] 
        for prod in self.products:
            for dev in prod.devices:
                result.append(dev.address)
        emb = self.gw.get_attribute('embedded')
        emb.value = result
        
    def get_value(self,node_id,cmd_class,instance=1,idx=0):
        node = self.network.nodes[node_id]
        for k in node.values:
            val = node.values[k]
            # some devices have broken command_class
            if val.command_class == None: continue
            if ((cmd_class == cmdclass.COMMAND_CLASS(val.command_class)) and (val.index==idx) and (val.instance == instance)):
                return val
        return None

    def _exit(self):
        if self.save_config:
            logger.info('Saving configuration file')
            self.cfg.write()
    
    def dump_product(self,node_id):
        """ dumb method that display a zwave device"""
        zdev = self.network.nodes[node_id]
        print("***** %s" % zdev.product_name)
        print("***** %s:%s:%s" % (zdev.manufacturer_id,zdev.product_id,zdev.product_type))
        table = PrettyTable(["value","inst","idx","Label","Data","Units","Command class"])
        table.align["Label"] = 'l'
        table.align["Data"] = 'l'
        table.align["Command class"] = 'l'
        for k in zdev.values:
            val = zdev.values[k]
            if val.command_class:
                klass = cmdclass.COMMAND_CLASS(val.command_class)
            else:
                klass = 'Error'
            table.add_row([k,val.instance,val.index,val.label,val.data,val.units,klass])
        print(table)
        
        
def setup(eng):
    try:
        gw=GW(eng)
        return True
    except ZWaveException as err:
        logger.warning('Error w/ Zwave network: %s' % err.value)




import openzwave
from openzwave.node import ZWaveNode
from openzwave.value import ZWaveValue
from openzwave.scene import ZWaveScene
from openzwave.controller import ZWaveController
from openzwave.network import ZWaveNetwork
from openzwave.option import ZWaveOption
from pydispatch import dispatcher
import sys
import time

from .cmdclass import COMMAND_CLASS
#import auto

device="/dev/zwave"
network = None

def connected():
    print("Connecting..")

def ready():
    print("Ready")

def disconnected():
    print("Disconnected")

def node_update(network, node):
    print('signal : Node update : {}.'.format(node))

def value_update(network, node, value):
    print('signal : Value update : {}.'.format(value))


def dump_device(node_id):
    zdev = network.nodes[node_id]
    print("***** %s" % zdev.product_name)
    print("***** %s:%s:%s" % (zdev.manufacturer_id,zdev.product_id,zdev.product_type))
    for k in zdev.values:
        val = zdev.values[k]
        klass = None
        if val.command_class !=None:
            klass = COMMAND_CLASS(val.command_class)
        print("%s/%s/%s [%s] %s%s %s" % (k,val.instance,val.index,val.label,val.data,val.units,klass))


def get_value(node_id,cmd_class,instance=1,idx=0):
    zdev = network.nodes[node_id]
    for k in zdev.values:
        val = zdev.values[k]
        if ((cmd_class == COMMAND_CLASS(val.command_class)) and (val.index==idx) and (val.instance == instance)):
            return val
    return None


options = ZWaveOption(device)
options.set_console_output(False)
options.lock()

network = ZWaveNetwork(options, autostart=False)
network.start()

dispatcher.connect(connected, ZWaveNetwork.SIGNAL_NETWORK_STARTED)
dispatcher.connect(disconnected, ZWaveNetwork.SIGNAL_NETWORK_FAILED)
dispatcher.connect(ready, ZWaveNetwork.SIGNAL_NETWORK_READY)
dispatcher.connect(node_update, ZWaveNetwork.SIGNAL_NODE)



for i in range(0,90):
    if network.is_ready:
        print("***** Network is ready")
        dispatcher.connect(value_update, ZWaveNetwork.SIGNAL_VALUE)
        break
    else:
        sys.stdout.write(".")
        sys.stdout.flush()
        time.sleep(1.0)

#dump_device(2)


#zdev = network.nodes[10]
#values = zdev.values

import pdb;pdb.set_trace()






from xaal.lib import Device,tools
from .cmdclass import COMMAND_CLASS

class ZDevice(object):
    
    def __init__(self,node,gateway):
        self.node = node
        self.gw = gateway
        self.devices = []
        self.values = {}
        self.config()
        self.setup()
        
    def config(self):
        """ read config files, search base_addr & group"""
        cfg = self.gw.cfg['products']
        key = str(self.node.node_id)
        if key in cfg.keys():
            tmp = cfg[key]['base_addr']
            self.base_addr =  tools.get_uuid(tmp)
        else:
            self.gw.save_config=True
            cfg.update({key:{}})
            cfg[key]['base_addr'] = str(tools.get_random_base_uuid())
            cfg[key]['group'] = str(tools.get_random_uuid())
        cfg.inline_comments[key] = self.node.product_name

    def new_device(self,devtype):
        """ embed an new device """
        node_id = str(self.node.node_id)
        cfg = self.gw.cfg['products'][node_id]
        dev = Device(devtype)
        dev.vendor_id  = "IHSEV/OpenZWave"
        dev.product_id = self.node.product_name
        dev.hw_id      = node_id
        dev.url        = "http://www.openzwave.com"
        dev.address    = tools.get_uuid(cfg['base_addr']) + len(self.devices) + 1
        dev.group_id   = tools.get_uuid(cfg['group'])
        dev.info       = '%s/%s' % (self.node.type,node_id)
        self.devices.append(dev)
        return dev

    def start(self):
        for k in self.values:
            self.handle_value_changed(self.values[k])
            self.values[k].refresh()
            
    def monitor_value(self,name,cmd_class,instance=1,idx=0):
        val = self.gw.get_value(self.node.node_id,cmd_class,instance,idx)
        if val:
            self.values[name] = val

    def set_value(self,name,val):
        self.values[name].data = val

    def get_value(self,name):
        return self.values[name]
        
    def dump(self):
        self.gw.dump_product(self.node.node_id)
    
    def setup(self):
        """ You should overide this for your product"""
        pass
    
    def handle_value_changed(self,value):
        """ You can overide this for your product"""
        pass
    
        
